;------------------------------------------------------------------------------
; CAD Concepts Limited
; 
; CONVERT2BLOCK
; 
; Copyright (C) 2022 CAD Concepts Limited.

;------------------------------------------------------------------------------
; File    : Convert2Block.LSP
; Author  : Jason Bourhill
; Web			: http://www.cadconcepts.co.nz
; Date    : 06/June/2022
;
; PURPOSE:
; Converts selected entities into a block(s)
;
; USAGE:
; To load type (load "CONVERT2BLOCK.LSP") from the command line or drag and drop the
; file onto your drawing using explorer.
;
; COMMAND(S)
;
; CONVERT2BLOCK
; Creates a single block from the selected objects.
; 
; CONVERT2BLOCKS
; Create individual blocks of each of the selected entities selected.
;
; REQUIRES: 
(vl-load-com)  ; Make sure the visual lisp extensions are available. Required by AutoCAD.
(load "VLE-Extension.lsp")            ; Make sure BricsCAD VLE Emulation is available. Required by AutoCAD, and portentially earlier versions of BricsCAD
;------------------------------------------------------------------------------

; GET-CURRENTUCS
; returns the current UCS as pointlist in the form
; '((origin) (xdirection vector) (ydirection vector)

(defun get-currentucs ()
  (list
    (getvar 'UCSORG)   ;origin
    (getvar 'UCSXDIR)  ;xvector
    (getvar 'UCSYDIR)  ;yvector
  )
)

; WORLDUCS-P
; Predicate to check whether the current UCS is the WORLD UCS.
; Returns T if true, nill otherwise

(defun worlducs-p ( / worlducs)
  (setq worlducs '((0.0 0.0 0.0)(1.0 0.0 0.0)(0.0 1.0 0.0)))
  (not (vle-list-diff  worlducs  (GET-CURRENTUCS)))
)

; SET-UCS
; Sets the current UCS from a point list in the form
; '((origin) (xdirection vector) (ydirection vector))
; e.g. '((-58.0 0.0 147.0) (1.0 0.0 0.0) (0.0 0.0 1.0))
; If nil is passed instead of a points list, then it will
; it will set to the World Ucs

(defun set-ucs (pointlist / orig xdir ydir)
  (acet-sysvar-set '("CMDECHO" 0 "OSMODE" 0))
  (cond (pointlist
          (setq orig (car pointlist))
          (setq xdir (mapcar '+ orig (cadr pointlist)))
          (setq ydir (mapcar '+ orig (caddr pointlist)))
          (command "._UCS" "_3P" orig xdir ydir)
        )
        (T (command "._UCS" "_W"))
  )
  (acet-sysvar-restore)
)

; Create-Block
; (create-block <block name> [<Selection Set>/<ename>/<vla-object>] <Insertion Point>)
; Function allows you to pass a selection set, ename, or vla-object
; If the insertion point is nil, then the bottom left bounding box coord will be used

(defun create-block (blockname sset-enm ip / oList odoc oblock oArray)
	;Create object list based on the selection type passed
  (cond
    ((vle-picksetp sset-enm) (setq oList (mapcar 'vlax-ename->vla-object (vle-selectionset->list sset-enm))))
    ((vle-enamep sset-enm) (setq oList (vlax-ename->vla-object sset-enm)))
    ((vle-vlaobjectp sset-enm) (setq oList (list sset-enm)) (setq sset-enm (vlax-vla-object->ename sset-enm)))
  )
  (cond
    ((and (snvalid blockname) oList)  ; if the block name and object list is valid
      ;Check IP
      (if (not ip)                    ; If the ip isn't provided, then
        (setq ip (car (vle-getgeomextents sset-enm)))  ; set ip as the min x,y,z, extents of a boundary box that encloses the selection
      )
       (setq odoc (vla-get-activedocument (vlax-get-acad-object)))  ; Retrieve active document
       (setq oblock (vla-add (vla-get-blocks odoc) (vlax-3d-point ip) blockname))  ; Create block
       (vla-put-units oblock (getvar 'insunits))  ; set it's insunits to match the current doc
       ;create and array of objects to copy into our block
       (setq oArray
             (vlax-safearray-fill
               (vlax-make-safearray
                 vlax-vbObject
                 (cons 0 (1- (length oList)))
               )
                                  oList
             )
       )
       (vla-copyobjects odoc oArray oblock)       ; copy objects into block definition
    )
  )
  (list oblock ip)  ; return block and ip point
)


; CONVERT2BLOCK
; Converts the selected objects into a single block with minimal prompting

(defun C:Convert2Block (/ sset odoc currentucs blkprefix blkname oBlock oSpace)
  (princ "\nSelect objects to convert to a block ")
  (setq sset (ssget))
  (cond
    (sset
      (setq odoc (vla-get-activedocument (vlax-get-acad-object)))  ; retrieve active document
      (vla-endundomark odoc)
      (vla-startundomark odoc)  ; set undo mark
      (cond  ;if World UCS isn't current
               ((not (WORLDUCS-P))
                     (setq currentucs (GET-CURRENTUCS))  ; save the current ucs
                     (SET-UCS nil)  ; set to world ucs
               )
      )
          (setq blkprefix "QBlock-")
          (setq *QBNUM* (if (not *QBNUM*) 1 *QBNUM*))
          ;Find unique block name
      (while (tblsearch "BLOCK" (setq blkname (strcat blkprefix (itoa *QBNUM*))))
        (setq *QBNUM* (1+ *QBNUM*))
      )
          (setq oBlock (CREATE-BLOCK blkname sset nil))  ; Create block of the sselection set
          (setq oSpace (if (< 1 (getvar 'CVPORT)) (vla-get-modelspace oDoc) (vla-get-paperspace oDoc)))  ; grab the current space
          (vla-InsertBlock oSpace (vlax-3d-point (cadr oBlock)) blkname 1 1 1 0)  ; Insert Block
          (mapcar 'vla-delete (mapcar 'vlax-ename->vla-object (vle-selectionset->list sset)))  ; delete the original entities
          (if currentucs (SET-UCS currentucs)); if the current UCS was changed, then reset it
          (vla-endundomark odoc)  ; close undo mark
          (princ (strcat "\nCreated block: " blkname))
    )
    (T (princ "\nNothing selected "))
  )
  (prin1)
)

; CONVERT2BLOCKS
; Converts each of the selected entities into a block with minimal prompting

(defun C:Convert2Blocks (/ sset odoc blkprefix oSpace obj blkname oBlock blklist)
  (princ "\nSelect objects to convert to blocks ")
  (setq sset (ssget))
  (cond
    (sset
      (setq odoc (vla-get-activedocument (vlax-get-acad-object)))  ; retrieve active document
      (vla-endundomark odoc)
      (vla-startundomark odoc)  ; set undo mark
      (cond  ;if World UCS isn't current
               ((not (WORLDUCS-P))
                     (setq currentucs (GET-CURRENTUCS))  ; save the current ucs
                     (SET-UCS nil)  ; set to world ucs
               )
      )
          (setq blkprefix "QBlock-")
          (setq *QBNUM* (if (not *QBNUM*) 1 *QBNUM*))
          (setq oSpace (if (< 1 (getvar 'CVPORT)) (vla-get-modelspace oDoc) (vla-get-paperspace oDoc)))  ; grab the current space
          (vle-start-transaction)
      (foreach obj (mapcar 'vlax-ename->vla-object (vle-selectionset->list sset))
        ;Find unique block name
        (while (tblsearch "BLOCK" (setq blkname (strcat blkprefix (itoa *QBNUM*))))
          (setq *QBNUM* (1+ *QBNUM*))
        )
        (setq oBlock (CREATE-BLOCK blkname obj nil))  ; Create block of the entity from our sselection set
        (vla-InsertBlock oSpace (vlax-3d-point (cadr oBlock)) blkname 1 1 1 0)  ; Insert Block
        (vla-delete obj)  ; delete the original entities
        (setq blklist (append blklist (list blkname)))
      )
          (vle-end-transaction)
          (if currentucs (SET-UCS currentucs))  ; if the current UCS was changed, then reset it
          (vla-endundomark odoc)  ; close undo mark
          (princ (strcat "\nCreated " (itoa (length blklist)) " Blocks, from " (car blklist) " --> " (nth (1- (length blklist)) blklist)))
    )
    (T (princ "\nNothing selected "))
  )
  (prin1)
)

(princ "\nConvert to Blocks loaded, Command(s): CONVERT2BLOCK, CONVERT2BLOCKS")
(princ)
